
/*
 * Project      [tie]
 * Author       TriX
 * Description  Analyses the ROM (Getting of all subroutines by finding of all
 *              jsr  instructions
 */

#include <stdlib.h>
#include <stdio.h>

int ReadFile (char* File, long *filesize, int offset, unsigned char **Pointer)
{
   FILE *tmp;
   int i;

   if ((tmp = fopen (File, "rb")) == NULL)
   {
     printf ("Cant open file %s\n", File);
     exit (-1);
   }
   fseek (tmp, 0, SEEK_END);
   *filesize = ftell (tmp);
   fseek (tmp, offset, SEEK_SET);
   *Pointer = (unsigned char*) malloc (*filesize);
   i = 0;
   while (!feof (tmp))
   {
      (*Pointer)[i] = fgetc (tmp);
      i++;
   } /* while */
   fclose (tmp);
   return 1;

} /* ReadFile */

int hexwrite (FILE *outfile, char what)
{
  char table[] = "0123456789abcdef";
  fputc (table[(what >> 4) & 0xf], outfile);
  fputc (table[what & 0xf], outfile);
  return 1;
}

/* This functions is the help functions for the quicksort algorithm */
int SortROMfunctions (const void *stuff_1, const void *stuff_2)
{
   unsigned char *EL_1 = (char *) stuff_1, *EL_2 = (char *) stuff_2;
   long value_1, value_2;
   value_1 = (EL_1[0] << 24) + (EL_1[1] << 16) + (EL_1[2] << 8) + (EL_1[3]);
   value_2 = (EL_2[0] << 24) + (EL_2[1] << 16) + (EL_2[2] << 8) + (EL_2[3]);
   if (value_2 > value_1) {
      return -1;
   } else if (value_2 < value_1) {
      return 1;
   } else return 0;
} /* SortROMfunctions */

int main (void)
{
   unsigned char *TI92ROM;
   long ROMSize;
   long position;
   unsigned char NULLMarker[] = {0, 0, 0, 0};
   unsigned char JSRMarker[] = {0x4e, 0xb9, 0x00};
   unsigned char RTSMarker[] = {0x4e, 0x75};
   unsigned char ROMAddresses[0xf00*4] = {0};
   long i;
   FILE *output;

   ReadFile ("romdump", &ROMSize, 0, &TI92ROM);
   /* Find JSR instructions */
   position = 0;
   while (position < ROMSize)
   {
      while (memcmp (TI92ROM + position, JSRMarker, 3) != 0 && position < ROMSize)
         position++;
      printf ("\r%ld/%ld", position, ROMSize);
      if (((TI92ROM[position + 3] >> 4) & 0xf) < 4)
      {
         position += 6;
         continue;
      }
      if (position == ROMSize) break;
      for (i = 0; i < 0xf00; i++)
      {
         if (memcmp (TI92ROM + position + 2, &ROMAddresses[4 * i], 4) == 0) break;
         if (memcmp (&ROMAddresses[4 * i], NULLMarker, 4) == 0)
         {
            ROMAddresses[4 * i] = TI92ROM[position + 2];
            ROMAddresses[4 * i + 1] = TI92ROM[position + 3];
            ROMAddresses[4 * i + 2] = TI92ROM[position + 4];
            ROMAddresses[4 * i + 3] = TI92ROM[position + 5];
            break;
         } /* if */
      } /* for */
      position += 6;
      while (memcmp (TI92ROM + position, RTSMarker, 2) != 0) position++;
      position += 2;
   } /* while */
   for (i = 0; i < 0xf00; i++)
   {
      if (memcmp (&ROMAddresses[4 * i], NULLMarker, 4) == 0) break;
   } /* for */

   qsort (ROMAddresses, i, 4, SortROMfunctions);

   output = fopen ("romfuncs.s", "w");
   fprintf (output, "\n\nROM\tdc.l 0");
   for (i = 0; i < 0xf00; i++)
   {
      if (memcmp (&ROMAddresses[4 * i], NULLMarker, 4) == 0) break;
      fprintf (output, ",\n\tdc.l $");
      hexwrite (output, ROMAddresses[4 * i]);
      hexwrite (output, ROMAddresses[4 * i + 1]);
      hexwrite (output, ROMAddresses[4 * i + 2]);
      hexwrite (output, ROMAddresses[4 * i + 3]);
   } /* for */
   fprintf (output, "\n\n\tEND");
   printf ("\nNumber of functions found: %d\n", i + 1);
   fclose (output);
} /* main */

